/// <reference path="./types.d.ts" />
import './redirect';
import fs from 'node:fs';
import type { Rule } from 'eslint';
import type { ESLintUtils } from '@typescript-eslint/utils';
import type { ESLintConfig, Rules } from 'eslint-define-config';
// import findCacheDirectory from 'find-cache-dir';
import { typescriptRules } from './presets/typescript';
import { unicornRules } from './presets/unicorn';
import { eslintRules } from './presets/eslint';
import { reactRules } from './presets/react';
import { importRules } from './presets/import';
import { jsDocRules } from './presets/jsdoc';
import { graphqlRules } from './presets/graphql';
import { localRules } from './presets/local';
import { error, warn, off } from './constants';
import { tailwindRules } from './presets/tailwind';

export { error, warn, off };

declare global {
  interface Array<T> {
    filter(
      predicate: BooleanConstructor,
    ): Exclude<T, null | undefined | false | '' | 0>[];
  }
}

const unique = (...arr: (false | undefined | string | string[])[]): string[] => [
  ...new Set(arr.flat(1).filter(Boolean)),
];

const ensureArray = <T>(value?: T | T[]): T[] =>
  value == null ? [] : Array.isArray(value) ? value : [value];

type RuleLevel = 'error' | 'warn' | 'off' | 0 | 1 | 2;
type RuleEntry<Options> = RuleLevel | [RuleLevel, Partial<Options>];

declare module 'eslint-define-config/src/rules/react/no-unknown-property.d.ts' {
  export interface NoUnknownPropertyOption {
    extends: ('next' | 'emotion')[];
  }
}

export interface LocalRuleOptions {
  /** Bans import from the specifier '.' and '..' and replaces it with '.+/index' */
  'rules/no-import-dot': RuleEntry<unknown>;
  /**
   * Enforce template literal expressions to be of `string` type
   * @see [restrict-template-expressions](https://typescript-eslint.io/rules/restrict-template-expressions)
   */
  'rules/restrict-template-expressions': RuleEntry<{ allow: string[] }>;
  /** Ban assignment of empty object literals `{}` and replace them with `Object.create(null)` */
  'rules/no-empty-object-literal': RuleEntry<unknown>;
}

export type RuleOptions = Rules & Partial<LocalRuleOptions>;

/**
 * ESLint Configuration.
 * @see [ESLint Configuration](https://eslint.org/docs/latest/user-guide/configuring/)
 */
type Config = Omit<ESLintConfig, 'rules'> & {
  /**
   * Rules.
   * @see [Rules](https://eslint.org/docs/latest/user-guide/configuring/rules)
   */
  rules?: RuleOptions;

  /**
   */
  customRules?: {
    rule: () => Promise<{
      default: Rule.RuleModule | ESLintUtils.RuleModule<string, unknown[]>;
    }>;
    options?: RuleLevel;
  }[];
};

export function defineCustomRule<Options extends readonly unknown[]>(
  rule: () => Promise<{
    default: Rule.RuleModule | ESLintUtils.RuleModule<string, Options>;
  }>,
  options?: Options,
) {
  return { rule, options };
}

/**
 * Returns a ESLint config object.
 *
 * By default, it includes `["@typescript-eslint", "import", "prettier"]` configs.
 * Additional bundled plugins include:
 *
 * 1. [`react`](https://github.com/jsx-eslint/eslint-plugin-react#list-of-supported-rules)
 *    (automatically enables
 *    [`react-hooks`](https://github.com/facebook/react/tree/main/packages/eslint-plugin-react-hooks))
 * 2. [`react-refresh`](https://github.com/ArnaudBarre/eslint-plugin-react-refresh)
 * 3. [`jsx-a11y`](https://github.com/jsx-eslint/eslint-plugin-jsx-a11y#supported-rules)
 * 4. [`unicorn`](https://github.com/sindresorhus/eslint-plugin-unicorn#rules)
 * 5. [`n`](https://github.com/eslint-community/eslint-plugin-n#-rules) (Node.js specific)
 * 6. [`jsdoc`](https://github.com/gajus/eslint-plugin-jsdoc#rules)
 *
 * Non bundled:
 * 1. [`graphql`](https://the-guild.dev/graphql/eslint/rules)
 */
export function extendConfig({
  plugins = [],
  settings,
  rules,
  extends: _extends,
  overrides,
  customRules,
  ...rest
}: Config = {}): ESLintConfig {
  const hasReact = plugins.includes('react');
  const hasReactRefresh = plugins.includes('react-refresh');
  const hasUnicorn = plugins.includes('unicorn');
  const hasJsDoc = plugins.includes('jsdoc');
  const hasGraphQL = plugins.includes('@graphql-eslint');
  const hasNext = ensureArray(_extends).some(name => name.includes(':@next/next'));
  const hasTailwind = ensureArray(_extends).some(name =>
    name.includes('plugin:tailwindcss/'),
  );

  const ruleDir = false; // ?? findCacheDirectory({ name: '_eslint-rules' });
  if (ruleDir) {
    fs.rmSync(ruleDir, { recursive: true, force: true });
    fs.mkdirSync(ruleDir, { recursive: true });
  }

  const result: Config = {
    root: true,
    parser: '@typescript-eslint/parser',
    plugins: unique('@typescript-eslint', 'import', 'rules', plugins),
    env: { node: true, browser: true, es2023: true },
    reportUnusedDisableDirectives: true,
    parserOptions: {
      project: true,
    },
    extends: unique(
      'eslint:recommended',
      'prettier',
      'plugin:@typescript-eslint/recommended-type-checked',
      'plugin:import/errors',
      'plugin:import/typescript',
      hasReact && [
        'plugin:react/recommended',
        'plugin:react-hooks/recommended',
        'plugin:jsx-a11y/recommended',
      ],
      hasJsDoc && 'plugin:jsdoc/recommended-typescript',
      hasGraphQL && 'plugin:@graphql-eslint/recommended',
      _extends,
    ),
    settings: {
      'import/parsers': {
        '@typescript-eslint/parser': ['.ts', '.tsx', '.mts', '.cts'],
      },
      'import/resolver': {
        typescript: {
          alwaysTryTypes: true,
        },
      },
      react: {
        version: 'detect',
      },
      ...settings,
    },
    overrides: [
      {
        files: [
          '.eslintrc.js',
          '*.config.js',
          'index.js',
          'babel.config.js',
          'next.config.js',
        ],
        extends: ['plugin:@typescript-eslint/disable-type-checked'],
        rules: {
          'rules/restrict-template-expressions': off,
        },
      },
      {
        files: ['*.d.ts'],
        rules: {
          '@typescript-eslint/consistent-type-imports': off,
        },
      },
      {
        files: ['repl.ts'],
        rules: {
          'no-console': off,
        },
      },
      ...(overrides ?? []),
    ],
    rules: {
      ...eslintRules,
      ...typescriptRules,
      ...importRules,
      ...localRules,
      ...(hasReact && {
        ...reactRules,
        'react/no-unknown-property': [
          error,
          { ignore: hasNext ? ['css', 'next'] : ['css'] },
        ],
      }),
      ...(hasReactRefresh && {
        'react-refresh/only-export-components': [warn, { allowConstantExport: true }],
      }),
      ...(hasUnicorn && unicornRules),
      ...(hasJsDoc && jsDocRules),
      ...(hasGraphQL && graphqlRules),
      ...(hasTailwind && tailwindRules),
      ...rules,
    },
    ...rest,
  };

  return result;
}
